import asyncio
from botstory.ast import story_context
import logging
from . import validate
from .. import commonhttp
from ... import di
from ...middlewares import option

logger = logging.getLogger(__name__)


@di.desc('fb', reg=False)
class FBInterface:
    type = 'facebook'

    def __init__(self,
                 api_uri='https://graph.facebook.com/v2.6',
                 greeting_text=None,
                 page_access_token='?',
                 persistent_menu=None,
                 webhook_url=None,
                 webhook_token=None,
                 ):
        """

        :param api_uri:
        :param greeting_text:
        :param page_access_token:
        :param persistent_menu:
        :param webhook_url:
        :param webhook_token:
        """
        self.api_uri = api_uri
        self.greeting_text = greeting_text
        self.persistent_menu = persistent_menu
        self.token = page_access_token
        self.webhook = webhook_url
        self.webhook_token = webhook_token

        self.library = None
        self.http = None
        self.story_processor = None
        self.storage = None
        self.users = None

    @di.inject()
    def add_library(self, stories_library):
        logger.debug('add_library')
        logger.debug(stories_library)
        self.library = stories_library

    @di.inject()
    def add_http(self, http):
        """
        inject http provider

        :param http:
        :return:
        """
        logger.debug('add_http')
        logger.debug(http)
        self.http = http

    @di.inject()
    def add_processor(self, story_processor):
        logger.debug('add_processor')
        logger.debug(story_processor)
        self.story_processor = story_processor

    @di.inject()
    def add_storage(self, storage):
        logger.debug('add_storage')
        logger.debug(storage)
        self.storage = storage

    @di.inject()
    def add_users(self, users):
        logger.debug('add_users')
        logger.debug(users)
        self.users = users

    async def send_text_message(self, recipient, text,
                                quick_replies=None,
                                options=None):
        """
        async send message to the facebook user (recipient)

        :param recipient:
        :param text:
        :param quick_replies:
        :param options

        :return:
        """

        try:
            validate.send_text_message(text, quick_replies)
        except validate.ExceedLengthException as i:
            # TODO: take first part of message show option `more`
            # store last part until user press `more`

            # TODO: register dynamic `quick answer` handler
            # with the rest of message

            # TODO: or handle it on application level?
            # motivation: if we're working with dynamic data or endless
            # we could just pass text. It should be generator
            logger.warn(str(i))

            if options and options.get('overflow', None) == 'cut':
                text = text[:i.limit]
            else:
                text = text[:i.limit - 2] + '\u2026'
        except validate.Invalid as i:
            logger.warn(str(i))

        if not quick_replies:
            quick_replies = []

        message = {
            'text': text,
        }

        quick_replies = [{**reply, 'content_type': 'text'} for reply in quick_replies]
        if len(quick_replies) > 0:
            message['quick_replies'] = quick_replies

        return await self.http.post(
            self.api_uri + '/me/messages/',
            params={
                'access_token': self.token,
            },
            json={
                'recipient': {
                    'id': recipient['facebook_user_id'],
                },
                'message': message,
            })

    async def request_profile(self, facebook_user_id):
        """
        Make request to facebook
        to receive more information about user

        More: https://developers.facebook.com/docs/messenger-platform/user-profile

        :param facebook_user_id:
        :return:
        """
        return await self.http.get(
            '{}/{}'.format(self.api_uri, facebook_user_id),
            params={
                'access_token': self.token,
            },
        )

    async def handle(self, data):
        logger.debug('')
        logger.debug('> handle <')
        logger.debug('')
        logger.debug('  entry: {}'.format(data))
        try:
            for e in data.get('entry', []):
                messaging = e.get('messaging', [])
                logger.debug('  messaging: {}'.format(messaging))

                if len(messaging) == 0:
                    logger.warning('  entry {} list lack of "messaging" field'.format(e))

                for m in messaging:
                    logger.debug('  m: {}'.format(m))

                    facebook_user_id = m['sender']['id']

                    logger.debug('before get user with facebook_user_id={}'.format(facebook_user_id))
                    user = await self.storage.get_user(facebook_user_id=facebook_user_id)
                    if not user:
                        logger.debug('  should create new user {}'.format(facebook_user_id))

                        try:
                            messenger_profile_data = await self.request_profile(facebook_user_id)
                            logger.debug('receive fb profile {}'.format(messenger_profile_data))
                        except commonhttp.errors.HttpRequestError as err:
                            logger.debug('fail on request fb profile of {}. with {}'.format(facebook_user_id, err))
                            messenger_profile_data = {
                                'no_fb_profile': True,
                            }

                        logger.debug('before creating new user')
                        user = await self.storage.new_user(
                            facebook_user_id=facebook_user_id,
                            no_fb_profile=messenger_profile_data.get('no_fb_profile', None),
                            first_name=messenger_profile_data.get('first_name', None),
                            last_name=messenger_profile_data.get('last_name', None),
                            profile_pic=messenger_profile_data.get('profile_pic', None),
                            locale=messenger_profile_data.get('locale', None),
                            timezone=messenger_profile_data.get('timezone', None),
                            gender=messenger_profile_data.get('gender', None),
                        )

                        self.users.on_new_user_comes(user)

                    session = await self.storage.get_session(facebook_user_id=facebook_user_id)
                    if not session:
                        logger.debug('  should create new session for user {}'.format(facebook_user_id))
                        session = await self.storage.new_session(
                            facebook_user_id=facebook_user_id,
                            stack=[],
                            user=user,
                        )

                    ctx = story_context.clean_message_data({
                        'session': session,
                        'user': user,
                    })

                    if 'message' in m:
                        logger.debug('message notification')
                        raw_message = m.get('message', {})
                        if 'is_echo' in raw_message:
                            # TODO: should react somehow.
                            # for example storing for debug purpose
                            logger.debug('just echo message')
                        else:
                            text = raw_message.get('text', None)
                            if text is not None:
                                ctx = story_context.set_message_data(ctx,
                                                                     'text', {
                                                                         'raw': text,
                                                                     })
                            elif 'sticker_id' in raw_message:
                                ctx = story_context.set_message_data(ctx,
                                                                     'sticker_id', raw_message['sticker_id'],
                                                                     )
                            else:
                                logger.warning('  entry {} "text"'.format(e))

                            quick_reply = raw_message.get('quick_reply', None)
                            if quick_reply is not None:
                                ctx = story_context.set_message_data(ctx,
                                                                     'option', quick_reply['payload'])

                            ctx = await self.story_processor.match_message(ctx)

                    elif 'postback' in m:
                        ctx = story_context.set_message_data(ctx,
                                                             'option', m['postback']['payload'])
                        ctx = await self.story_processor.match_message(ctx)
                    elif 'delivery' in m:
                        logger.debug('delivery notification')
                    elif 'read' in m:
                        logger.debug('read notification')
                    else:
                        logger.warning('(!) unknown case {}'.format(e))

                    # after message were processed session and user information could change
                    # so we should store it for the next usage
                    await self.storage.set_session(ctx['session'])
                    await self.storage.set_user(ctx['user'])

        except BaseException as err:
            logger.exception(err)

        return {
            'status': 200,
            'text': 'Ok!',
        }

    async def setup(self):
        logger.debug('setup')

        if self.greeting_text:
            asyncio.ensure_future(
                self.replace_greeting_text(self.greeting_text)
            )

        if self.persistent_menu:
            asyncio.ensure_future(
                self.replace_persistent_menu(self.persistent_menu)
            )

        # check whether we have `On Start Story`
        have_on_start_story = not not self.library.get_global_story(story_context.set_message_data({
            'session': {}
        }, 'option', option.OnStart.DEFAULT_OPTION_PAYLOAD))
        if have_on_start_story:
            await self.remove_greeting_call_to_action_payload()
            await self.set_greeting_call_to_action_payload(option.OnStart.DEFAULT_OPTION_PAYLOAD)

    async def before_start(self):
        logger.debug('start')
        if self.webhook and self.http:
            self.http.webhook(self.webhook, self.handle, self.webhook_token)

    async def replace_greeting_text(self, message):
        """
        delete greeting text before
        :param message:
        :return:
        """
        try:
            await self.remove_greeting_text()
        except Exception:
            pass

        await self.set_greeting_text(message)

    async def set_greeting_text(self, message):
        """
        set a greeting for new conversations

        can use for personalizing
            {{user_first_name}}
            {{user_last_name}}
            {{user_full_name}}

        more: https://developers.facebook.com/docs/messenger-platform/thread-settings/greeting-text

        :param message:
        :return:
        """
        logger.debug('set_greeting_text')
        try:
            validate.greeting_text(message)
        except validate.Invalid as i:
            logger.warn(str(i))

        self.greeting_text = message

        if not self.http:
            # should wait until receive http
            return

        await self.http.post(
            self.api_uri + '/me/thread_settings',
            params={
                'access_token': self.token,
            },
            json={
                'setting_type': 'greeting',
                'greeting': {
                    'text': message,
                },
            }
        )

    async def remove_greeting_text(self):
        logger.debug('remove_greeting_text')
        if not self.http:
            return

        await self.http.delete(
            self.api_uri + '/me/thread_settings',
            params={
                'access_token': self.token,
            },
            json={
                'setting_type': 'greeting',
            }
        )

    async def set_greeting_call_to_action_payload(self, payload):
        logger.debug('set_greeting_call_to_action_payload')
        """

        more: https://developers.facebook.com/docs/messenger-platform/thread-settings/get-started-button

        :param payload:
        :return:
        """
        await self.http.post(
            self.api_uri + '/me/thread_settings',
            params={
                'access_token': self.token,
            },
            json={
                'setting_type': 'call_to_actions',
                'thread_state': 'new_thread',
                'call_to_actions': [{'payload': payload}]
            }
        )

    async def remove_greeting_call_to_action_payload(self):
        logger.debug('remove_greeting_call_to_action_payload')
        if not self.http:
            return

        await self.http.delete(
            self.api_uri + '/me/thread_settings',
            params={
                'access_token': self.token,
            },
            json={
                'setting_type': 'call_to_actions',
                'thread_state': 'new_thread',
            }
        )

    async def replace_persistent_menu(self, menu):
        logger.debug('replace_persistent_menu')
        try:
            await self.remove_persistent_menu()
        except Exception:
            pass
        await self.set_persistent_menu(menu)

    async def set_persistent_menu(self, menu):
        """
        more: https://developers.facebook.com/docs/messenger-platform/thread-settings/persistent-menu

        :param menu:
        :return:
        """
        logger.debug('set_persistent_menu')
        try:
            validate.persistent_menu(menu)
        except validate.Invalid as i:
            logger.warn(str(i))

        self.persistent_menu = menu

        if not self.http:
            # should wait until receive http
            return

        await self.http.post(
            self.api_uri + '/me/thread_settings',
            params={
                'access_token': self.token,
            },
            json={
                'setting_type': 'call_to_actions',
                'thread_state': 'existing_thread',
                'call_to_actions': menu,
            }
        )

    async def remove_persistent_menu(self):
        logger.debug('remove_persistent_menu')
        if not self.http:
            return

        await self.http.delete(
            self.api_uri + '/me/thread_settings',
            params={
                'access_token': self.token,
            },
            json={
                'setting_type': 'call_to_actions',
                'thread_state': 'existing_thread',
            }
        )



from botstory.ast import story_context
from botstory.middlewares import text
import logging
import os

import pytest

from . import fake_server
from .. import aiohttp, fb, mongodb, mockhttp
from ... import Story, utils

logger = logging.getLogger(__name__)


story = None


def teardown_function(function):
    logger.debug('tear down!')
    story and story.clear()


def build_message(facebook_user_id, msg):
    return {
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [{
                'sender': {
                    'id': facebook_user_id,
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 1458692752478,
                'message': {
                    'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                    'seq': 73,
                    **msg,
                }
            }]
        }]
    }


@pytest.fixture
def build_context():
    async def builder(db, no_session=False, no_user=False):
        user = None
        if not no_user:
            user = utils.build_fake_user()
            await db.set_user(user)

        if not no_session:
            session = utils.build_fake_session(user=user)
            await db.set_session(session)

        global story
        story = Story()
        story.use(db)
        fb_interface = story.use(fb.FBInterface(page_access_token='qwerty'))
        http = story.use(mockhttp.MockHttpInterface())

        await story.start()

        return fb_interface, http, story, user

    return builder


@pytest.fixture
@pytest.mark.asyncio
def open_db():
    class AsyncDBConnection:
        def __init__(self):
            self.db_interface = mongodb.MongodbInterface(uri=os.environ.get('TEST_MONGODB_URL', 'mongo'),
                                                         db_name='test')

        async def __aenter__(self):
            await self.db_interface.start()
            await self.db_interface.clear_collections()
            return self.db_interface

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            self.db_interface.stop()
            await self.db_interface.clear_collections()
            self.db_interface = None

    return AsyncDBConnection


@pytest.mark.asyncio
async def test_facebook_interface_should_use_aiohttp_to_post_message(event_loop):
    async with fake_server.fake_fb.FakeFacebook(event_loop) as server:
        async with server.session() as server_session:
            # 1) setup app

            global story
            story = Story()

            try:
                story.use(fb.FBInterface(
                    webhook_url='/webhook',
                ))
                http_integration = story.use(aiohttp.AioHttpInterface())
                # TODO: could be better. Mock http interface
                http_integration.session = server_session

                await story.start()

                user = utils.build_fake_user()

                await story.say('Pryvit!', user=user)

                assert len(server.history) > 0
                req = server.history[-1]['request']
                assert req.content_type == 'application/json'
                assert await req.json() == {
                    'recipient': {
                        'id': user['facebook_user_id'],
                    },
                    'message': {
                        'text': 'Pryvit!'
                    }
                }
            finally:
                await story.stop()


@pytest.mark.asyncio
async def test_integrate_mongodb_with_facebook(open_db, build_context):
    async with open_db() as mongodb:
        facebook, _, story, user = await build_context(mongodb)

        trigger = utils.SimpleTrigger()

        @story.on('hello, world!')
        def correct_story():
            @story.part()
            def store_result(ctx):
                trigger.receive(ctx)

        await facebook.handle(build_message(
            user['facebook_user_id'], {
                'text': 'hello, world!'
            }
        ))

        assert trigger.value['user'] == user
        assert story_context.get_message_data(trigger.value) == {
            'text': {
                'raw': 'hello, world!'
            }
        }


@pytest.mark.asyncio
async def test_integrate_mongodb_with_facebook_with_none_session(open_db, build_context):
    async with open_db() as mongodb:
        facebook, _, story, _ = await build_context(mongodb, no_session=True, no_user=True)

        trigger = utils.SimpleTrigger()

        @story.on('hello, world!')
        def correct_story():
            @story.part()
            def store_result_for_new_user(message):
                trigger.receive(message)

        await facebook.handle(build_message('some-facebook-id', {
            'text': 'hello, world!'
        }))

        assert trigger.value
        assert text.get_text(trigger.value)['raw'] == 'hello, world!'
        assert trigger.value['user']['facebook_user_id'] == 'some-facebook-id'


@pytest.mark.asyncio
async def test_story_on_start(open_db, build_context):
    async with open_db() as mongodb:
        facebook, http, story, _ = await build_context(mongodb)

        trigger = utils.SimpleTrigger()

        @story.on_start()
        def just_meet():
            @story.part()
            def greeting(ctx):
                trigger.passed()

        await story.setup()

        http.delete.assert_called_with(
            'https://graph.facebook.com/v2.6/me/thread_settings',
            params={
                'access_token': 'qwerty',
            },
            json={
                'setting_type': 'call_to_actions',
                'thread_state': 'new_thread',
            }
        )

        http.post.assert_called_with(
            'https://graph.facebook.com/v2.6/me/thread_settings',
            params={
                'access_token': 'qwerty',
            },
            json={
                'setting_type': 'call_to_actions',
                'thread_state': 'new_thread',
                'call_to_actions': [
                    {
                        'payload': 'BOT_STORY.PUSH_GET_STARTED_BUTTON'
                    }
                ]
            }
        )

        await story.start()

        await facebook.handle({
            'object': 'page',
            'entry': [{
                'id': 'PAGE_ID',
                'time': 1473204787206,
                'messaging': [{
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'postback': {
                        'payload': 'BOT_STORY.PUSH_GET_STARTED_BUTTON'
                    }
                }]
            }]
        })

        assert trigger.is_triggered


# TODO:
@pytest.mark.skip()
@pytest.mark.asyncio
async def test_should_prevent_other_story_to_start_until_we_waiting_for_answer(open_db, build_context):
    async with open_db() as mongodb:
        facebook, http, story, user = await build_context(mongodb)

        trigger_1 = utils.SimpleTrigger()
        trigger_2 = utils.SimpleTrigger()

        @story.on('hi there!')
        def one_story():
            @story.part()
            async def then_ask(message):
                return await story.ask('How are you?', user=message['user'])

            @story.part()
            async def then_trigger_2(message):
                trigger_2.passed()

        @story.on('Great!')
        def another_story():
            @story.part()
            def then_trigger_1(message):
                trigger_1.passed()

        await story.start()

        await facebook.handle(build_message(user['facebook_user_id'], {
            'text': 'hi there!'
        }))
        await facebook.handle(build_message(user['facebook_user_id'], {
            'text': 'Great!'
        }))

        assert trigger_2.is_triggered
        assert not trigger_1.is_triggered


@pytest.mark.asyncio
async def test_should_not_setup_call_to_action_for_new_thread_if_we_dont_have_on_start(open_db, build_context):
    async with open_db() as mongodb:
        facebook, http, story, _ = await build_context(mongodb)

        @story.on('hi')
        def one_story():
            @story.part()
            def greeting(message):
                pass

        await story.start()

        # TODO: check whether we've told to fb that start button generate
        # `BOT_STORY.PUSH_GET_STARTED_BUTTON` payload
        assert not http.post.called, 'setup thread'



import pytest
from .. import di


def test_injector_get():
    with di.child_scope():
        di.injector.register('once_instance', 'Hello World!')
        assert di.injector.get('once_instance') == 'Hello World!'


def test_lazy_description_should_not_register_class():
    with di.child_scope():
        @di.desc(reg=False)
        class OneClass:
            pass

        assert di.injector.get('one_class') is None


def test_lazy_description_should_simplify_registration():
    with di.child_scope():
        @di.desc(reg=False)
        class OneClass:
            pass

        di.injector.register(instance=OneClass())

        assert isinstance(di.injector.get('one_class'), OneClass)


def test_not_lazy_description_should_simplify_registration():
    with di.child_scope():
        @di.desc(reg=True)
        class OneClass:
            pass

        assert isinstance(di.injector.get('one_class'), OneClass)


def test_fail_if_type_is_not_string():
    with di.child_scope():
        class OneClass:
            pass

        with pytest.raises(ValueError):
            di.injector.register(OneClass)


def test_kebab_string_style_is_synonym_to_underscore():
    with di.child_scope():
        @di.desc()
        class OneClass:
            pass

        assert isinstance(di.injector.get('one-class'), OneClass)


def test_later_binding():
    with di.child_scope():
        @di.desc()
        class OuterClass:
            @di.inject()
            def deps(self, test_class):
                self.test_class = test_class

        @di.desc('test_class', reg=False)
        class InnerClass:
            pass

        outer = OuterClass()
        di.injector.register(instance=outer)
        di.bind(outer, auto=True)

        inner = InnerClass()
        di.injector.register(instance=inner)

        assert outer.test_class == inner


def test_overwrite_previous_singleton_instance():
    with di.child_scope():
        @di.desc('test_class')
        class FirstClass:
            pass

        first_class = di.get('test_class')

        @di.desc('test_class')
        class SecondClass:
            pass

        second_class = di.get('test_class')

        assert first_class != second_class
        assert isinstance(first_class, FirstClass)
        assert isinstance(second_class, SecondClass)


def test_inherit_scope():
    with di.child_scope('first'):
        @di.desc()
        class First:
            pass

        with di.child_scope('second'):
            @di.desc()
            class Second:
                @di.inject()
                def deps(self, first):
                    self.first = first

            second = di.get('second')
            assert isinstance(second, Second)
            assert isinstance(second.first, First)


def test_do_not_call_deps_endpoint_before_we_have_all_needed_deps():
    with di.child_scope():
        @di.desc()
        class Container:
            def __init__(self):
                self.one = 'undefined'
                self.two = 'undefined'

            @di.inject()
            def deps(self, one, two):
                self.one = one
                self.two = two

        container = di.get('container')
        assert container.one == 'undefined'
        assert container.two == 'undefined'

        @di.desc()
        class One:
            pass

        di.bind(container)
        assert container.one == 'undefined'
        assert container.two == 'undefined'

        @di.desc()
        class Two:
            pass

        di.bind(container)
        assert isinstance(container.one, One)
        assert isinstance(container.two, Two)


def test_do_not_call_deps_endpoint_before_we_have_all_needed_deps_or_default_value():
    with di.child_scope():
        @di.desc()
        class Container:
            def __init__(self):
                self.one = 'undefined'
                self.two = 'undefined'

            @di.inject()
            def deps(self, one, two='default'):
                self.one = one
                self.two = two

        container = di.get('container')
        assert container.one == 'undefined'
        assert container.two == 'undefined'

        @di.desc()
        class One:
            pass

        di.bind(container)
        assert isinstance(container.one, One)
        assert container.two == 'default'

        @di.desc()
        class Two:
            pass

        di.bind(container)
        assert isinstance(container.one, One)
        assert isinstance(container.two, Two)



import botstory
from botstory.ast import story_context
from botstory.utils import build_fake_session, build_fake_user


class Talk:
    def __init__(self, session=None, user=None, story=None):
        if not user:
            user = build_fake_user()
        if not session:
            session = build_fake_session()
        self.session = session
        self.user = user
        self.story = story

    def __enter__(self):
        self.story = botstory.Story()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.story.clear()
        return False

    def __call__(self, *args, **kwargs):
        return self.wrap_user_talk(args[0])

    def init_by_ctx(self, ctx):
        self.session = ctx.message['session']

    async def location(self, loc):
        return await self.wrap_user_talk(location)(loc)

    async def pure_text(self, text):
        return await self.wrap_user_talk(pure_text)(text)

    async def option(self, payload):
        return await self.wrap_user_talk(option)(payload)

    async def ask(self, payload):
        if payload is None:
            return await self.story.match_message(
                story_context.set_message_data(
                    story_context.clean_message_data({
                        'session': self.session,
                        'user': self.user,
                    }), {}))
        if 'sticker_id' in payload:
            return await self.wrap_user_talk(sticker)(payload)
        raise NotImplementedError('put all other message types here')

    def wrap_user_talk(self, fn):
        async def fn_wrapper(payload):
            mutated_ctx = await fn(payload,
                                   self.session,
                                   self.user,
                                   self.story)
            self.session = mutated_ctx['session']
            self.user = mutated_ctx['user']
            return mutated_ctx

        return fn_wrapper


async def location(loc, session=None, user=None, story=None):
    return await story.match_message(
        story_context.set_message_data(
            story_context.clean_message_data({
                'session': session,
                'user': user,
            }), 'location', loc))


async def pure_text(text, session=None, user=None, story=None):
    return await story.match_message(
        story_context.set_message_data(
            story_context.clean_message_data({
                'session': session,
                'user': user,
            }), 'text', {
                'raw': text,
            }))


async def option(payload, session=None, user=None, story=None):
    return await story.match_message(
        story_context.set_message_data(
            story_context.clean_message_data({
                'session': session,
                'user': user,
            }), 'option', payload))


async def sticker(payload, session=None, user=None, story=None):
    return await story.match_message(
        story_context.set_message_data(
            story_context.clean_message_data({
                'session': session,
                'user': user,
            }), 'sticker_id', payload['sticker_id']))



import pytest

from botstory.ast import loop
from botstory.utils import answer, SimpleTrigger


@pytest.mark.asyncio
async def test_jump_in_a_loop():
    trigger_show_global_help = SimpleTrigger()
    trigger_show_local_help = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('?')
        def global_help_story():
            @story.part()
            def show_global_help(ctx):
                trigger_show_global_help.passed()

        @story.on('start job')
        def one_job():
            @story.part()
            def show_do_job(ctx):
                pass

            @story.loop()
            def job_scope():
                @story.on('?')
                def local_help_story():
                    @story.part()
                    def show_local_help(ctx):
                        trigger_show_local_help.passed()

                @story.on('work')
                def job_story():
                    @story.part()
                    def do_some_job(ctx):
                        pass

        await talk.pure_text('start job')
        await talk.pure_text('?')

    assert trigger_show_global_help.is_triggered is not True
    assert trigger_show_local_help.is_triggered is True


@pytest.mark.asyncio
async def test_do_not_propagate_previous_message_to_the_loop():
    trigger_hi_again = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('hi!')
        def one_job():
            @story.part()
            def show_do_job(ctx):
                pass

            @story.loop()
            def job_scope():
                @story.on('hi!')
                def hi_again():
                    @story.part()
                    def store_hi_again():
                        trigger_hi_again.passed()

                @story.on('?')
                def local_help_story():
                    @story.part()
                    def show_local_help(ctx):
                        pass

                @story.on('work')
                def job_story():
                    @story.part()
                    def do_some_job(ctx):
                        pass

        await talk.pure_text('hi!')

    assert trigger_hi_again.is_triggered is not True


@pytest.mark.asyncio
async def test_loop_inside_of_loop():
    trigger_global_action1 = SimpleTrigger()
    trigger_inner_action1 = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('action1')
        def global_action1():
            @story.part()
            def trigger_action1(ctx):
                trigger_global_action1.passed()

        @story.on('action2')
        def global_action2():
            @story.loop()
            def outer_loop():
                @story.on('action1')
                def outer_action1():
                    @story.loop()
                    def inner_loop():
                        @story.on('action1')
                        def inner_action1():
                            @story.part()
                            def inner_action1_part(ctx):
                                trigger_inner_action1.passed()

                        @story.on('action2')
                        def inner_action2():
                            @story.part()
                            def inner_action2_part(ctx):
                                pass

                @story.on('action2')
                def outer_action2():
                    @story.part()
                    def do_some_job(ctx):
                        pass

        await talk.pure_text('action2')
        await talk.pure_text('action1')
        await talk.pure_text('action1')

    assert trigger_inner_action1.is_triggered is True
    assert trigger_global_action1.is_triggered is not True


@pytest.mark.asyncio
async def test_looping_the_loop():
    trigger_show_global_help = SimpleTrigger()
    trigger_show_local_help = SimpleTrigger(0)

    with answer.Talk() as talk:
        story = talk.story

        @story.on('?')
        def global_help_story():
            @story.part()
            def show_global_help(ctx):
                trigger_show_global_help.passed()

        @story.on('start job')
        def one_job():
            @story.part()
            def show_do_job(ctx):
                pass

            @story.loop()
            def job_scope():
                @story.on('?')
                def local_help_story():
                    @story.part()
                    def show_local_help(ctx):
                        trigger_show_local_help.inc()

                @story.on('work')
                def job_story():
                    @story.part()
                    def do_some_job(ctx):
                        pass

        await talk.pure_text('start job')
        await talk.pure_text('?')
        await talk.pure_text('?')
        await talk.pure_text('?')

    assert trigger_show_local_help.value == 3
    assert trigger_show_global_help.is_triggered is not True


@pytest.mark.asyncio
async def test_breaking_the_loop():
    trigger_show_global_help = SimpleTrigger()
    trigger_show_local_help = SimpleTrigger(0)

    with answer.Talk() as talk:
        story = talk.story

        @story.on('?')
        def global_help_story():
            @story.part()
            def show_global_help(ctx):
                trigger_show_global_help.passed()

        @story.on('start job')
        def one_job():
            @story.part()
            def show_do_job(ctx):
                pass

            @story.loop()
            def job_scope():
                @story.on('?')
                def local_help_story():
                    @story.part()
                    def show_local_help(ctx):
                        trigger_show_local_help.inc()

                @story.on('work')
                def job_story():
                    @story.part()
                    def do_some_job(ctx):
                        return loop.BreakLoop()

        await talk.pure_text('start job')
        await talk.pure_text('?')
        await talk.pure_text('work')
        await talk.pure_text('?')

    assert trigger_show_local_help.value == 1
    assert trigger_show_global_help.is_triggered is True


@pytest.mark.asyncio
async def test_breaking_the_loop_inside_of_other_loop():
    trigger_global = SimpleTrigger()
    trigger_inner = SimpleTrigger()
    trigger_outer = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('action1')
        def global_action1():
            @story.part()
            def trigger_action1(ctx):
                pass

        @story.on('action2')  # 1
        def global_action2():
            @story.loop()
            def outer_loop():
                @story.on('action1')  # 2
                def outer_action1():
                    @story.loop()
                    def inner_loop():
                        @story.on('action1')  # 3
                        def inner_action1():
                            @story.part()
                            def inner_action1_part(ctx):
                                return loop.BreakLoop()

                        @story.on('action2')
                        def inner_action2():
                            @story.part()
                            def inner_action2_part(ctx):
                                pass

                        @story.on('action3')  # 4 (wrong)
                        def inner_action3():
                            @story.part()
                            def inner_action3_part(ctx):
                                trigger_inner.passed()

                @story.on('action2')
                def outer_action2():
                    @story.part()
                    def do_some_job(ctx):
                        pass

                @story.on('action3')  # 4 (correct)
                def outer_action3():
                    @story.part()
                    def do_some_job(ctx):
                        trigger_outer.passed()

            @story.on('action3')  # 4 (wrong)
            def global_action3():
                @story.part()
                def do_some_job(ctx):
                    trigger_global.passed()

        await talk.pure_text('action2')
        await talk.pure_text('action1')
        await talk.pure_text('action1')
        await talk.pure_text('action3')

    assert trigger_global.is_triggered is False
    assert trigger_inner.is_triggered is False
    assert trigger_outer.is_triggered is True


@pytest.mark.asyncio
async def test_prevent_message_propagation_from_outer_loop_to_an_inner_loop():
    trigger_1 = SimpleTrigger(0)
    trigger_2 = SimpleTrigger(0)
    trigger_3 = SimpleTrigger(0)

    with answer.Talk() as talk:
        story = talk.story

        @story.on('action1')  # 1
        def global_action1():
            @story.part()
            def pre_1(ctx):
                trigger_1.inc()

            @story.loop()
            def outer_loop():
                @story.on('action1')  # 2, 4
                def outer_action1():
                    @story.part()
                    def pre_2(ctx):
                        trigger_2.inc()

                    @story.loop()
                    def inner_loop():
                        @story.on('action1')  # 3
                        def inner_action1():
                            @story.part()
                            def inner_action1_part(ctx):
                                trigger_3.inc()
                                return loop.BreakLoop()

                    @story.part()
                    def pre_3(ctx):
                        pass

        await talk.pure_text('action1')
        await talk.pure_text('action1')
        await talk.pure_text('action1')
        await talk.pure_text('action1')

    assert trigger_1.value == 1
    assert trigger_2.value == 2
    assert trigger_3.value == 1


@pytest.mark.asyncio
async def test_break_loop_on_unmatched_message():
    action1_trigger = SimpleTrigger(0)
    action2_after_part_trigger = SimpleTrigger(0)

    with answer.Talk() as talk:
        story = talk.story

        @story.on('action1')
        def action1():
            @story.part()
            def action1_part(ctx):
                pass

        @story.on('action2')
        def action2():
            @story.part()
            def action2_part(ctx):
                pass

            @story.loop()
            def action2_loop():
                @story.on('action1')
                def action2_loop_action1():
                    @story.part()
                    def action2_loop_action1_part(ctx):
                        action1_trigger.inc()

                @story.on('action2')
                def action2_loop_action2():
                    @story.part()
                    def action2_loop_action2_part(ctx):
                        pass

            @story.part()
            def action2_after_part(ctx):
                action2_after_part_trigger.inc()

        await talk.pure_text('action2')
        await talk.pure_text('action1')
        await talk.pure_text('action1')
        await talk.pure_text('action3')

    assert action1_trigger.value == 2
    assert action2_after_part_trigger.value == 1


@pytest.mark.asyncio
async def test_break_loop_on_unmatched_message_and_jump_to_another_story():
    action1_trigger = SimpleTrigger(0)
    action3_trigger = SimpleTrigger(0)

    with answer.Talk() as talk:
        story = talk.story

        @story.on('action2')
        def action2():
            @story.part()
            def action2_part(ctx):
                pass

            @story.loop()
            def action2_loop():
                @story.on('action1')
                def action2_loop_action1():
                    @story.part()
                    def action2_loop_action1_part(ctx):
                        action1_trigger.inc()

                @story.on('action2')
                def action2_loop_action2():
                    @story.part()
                    def action2_loop_action2_part(ctx):
                        pass

        @story.on('action3')
        def action3():
            @story.part()
            def action3_part(ctx):
                action3_trigger.inc()

        await talk.pure_text('action2')
        await talk.pure_text('action1')
        await talk.pure_text('action1')
        await talk.pure_text('action3')

    assert action1_trigger.value == 2
    assert action3_trigger.value == 1



from botstory.ast import story_context
import json
import logging
import pytest
import random

from . import forking
from .. import matchers
from ..middlewares import location, text
from ..utils import answer, SimpleTrigger

logger = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_cases():
    trigger_location = SimpleTrigger()
    trigger_text = SimpleTrigger()
    trigger_after_switch = SimpleTrigger()

    with answer.Talk() as talk:
        say_location = talk(answer.location)
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('Hi there!')
        def one_story():
            @story.part()
            async def start(ctx):
                await story.say('Where do you go?', user=ctx['user'])
                return forking.Switch({
                    'location': location.Any(),
                    'text': text.Any(),
                })

            @story.case(match='location')
            def location_case():
                @story.part()
                def store_location(ctx):
                    trigger_location.receive(story_context.get_message_data(ctx))

            @story.case(match='text')
            def text_case():
                @story.part()
                def store_text(ctx):
                    trigger_text.receive(text.get_text(ctx)['raw'])

            @story.part()
            def after_switch(ctx):
                trigger_after_switch.passed()

        await say_pure_text('Hi there!')
        await say_location({'x': 123, 'y': 321})

        assert trigger_location.result() == {
            'location': {'x': 123, 'y': 321}
        }
        assert not trigger_text.result()
        assert trigger_after_switch.is_triggered


@pytest.mark.asyncio
async def test_sync_value():
    trigger_start = SimpleTrigger()
    trigger_heads = SimpleTrigger()
    trigger_tails = SimpleTrigger()

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('Flip a coin!')
        def one_story():
            @story.part()
            def start(message):
                coin = random.choice(['heads', 'tails'])
                assert not trigger_start.is_triggered
                assert not trigger_heads.is_triggered
                assert not trigger_tails.is_triggered
                trigger_start.passed()
                return forking.SwitchOnValue(coin)

            @story.case(equal_to='heads')
            def heads():
                @story.part()
                def store_heads(message):
                    assert not trigger_heads.is_triggered
                    trigger_heads.passed()

            @story.case(equal_to='tails')
            def tails():
                @story.part()
                def store_tails(message):
                    assert not trigger_tails.is_triggered
                    trigger_tails.passed()

        await say_pure_text('Flip a coin!')

        assert trigger_heads.is_triggered != trigger_tails.is_triggered


@pytest.mark.asyncio
async def test_few_switches_in_one_story():
    trigger_heads = SimpleTrigger()
    trigger_heads.receive(0)
    trigger_tails = SimpleTrigger()
    trigger_tails.receive(0)

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('Flip a coin!')
        def one_story():
            @story.part()
            def start_0(ctx):
                coin = random.choice(['heads', 'tails'])
                return forking.SwitchOnValue(coin)

            @story.case(equal_to='heads')
            def heads_0():
                @story.part()
                def store_heads(ctx):
                    trigger_heads.receive(trigger_heads.value + 1)

            @story.case(equal_to='tails')
            def tails_0():
                @story.part()
                def store_tails(ctx):
                    trigger_tails.receive(trigger_tails.value + 1)

            @story.part()
            def start_1(ctx):
                assert trigger_heads.value + trigger_tails.value == 1
                coin = random.choice(['heads', 'tails'])
                return forking.SwitchOnValue(coin)

            @story.case(equal_to='heads')
            def heads_1():
                @story.part()
                def store_heads(ctx):
                    trigger_heads.receive(trigger_heads.value + 1)

            @story.case(equal_to='tails')
            def tails_1():
                @story.part()
                def store_tails(ctx):
                    trigger_tails.receive(trigger_tails.value + 1)

        await say_pure_text('Flip a coin!')

        assert trigger_heads.value + trigger_tails.value == 2


@pytest.mark.asyncio
async def test_default_sync_value():
    trigger_1 = SimpleTrigger()
    trigger_default = SimpleTrigger()

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('Roll the dice!')
        def one_story_1():
            @story.part()
            def start(ctx):
                side = random.randint(1, 6)
                return forking.SwitchOnValue(side)

            @story.case(equal_to=1)
            def side_1():
                @story.part()
                def store_1(ctx):
                    trigger_1.passed()

            @story.case(default=True)
            def other_sides():
                @story.part()
                def store_other(ctx):
                    trigger_default.passed()

        await say_pure_text('Roll the dice!')

        assert trigger_1.is_triggered != trigger_default.is_triggered


@pytest.mark.asyncio
async def test_one_sync_switch_inside_of_another_sync_switch():
    visited_rooms = SimpleTrigger(0)

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('enter')
        def labyrinth():
            @story.part()
            def enter(ctx):
                turn = random.choice(['left', 'right'])
                visited_rooms.receive(visited_rooms.value + 1)
                return forking.SwitchOnValue(turn)

            @story.case(equal_to='left')
            def room_1():
                @story.part()
                def next_room_1(ctx):
                    turn = random.choice(['left', 'right'])
                    visited_rooms.receive(visited_rooms.value + 1)
                    return forking.SwitchOnValue(turn)

                @story.case(equal_to='left')
                def room_1_1():
                    @story.part()
                    def next_room_1_1(ctx):
                        visited_rooms.receive(visited_rooms.value + 1)

                @story.case(equal_to='right')
                def room_1_2():
                    @story.part()
                    def next_room_1_2(ctx):
                        visited_rooms.receive(visited_rooms.value + 1)

            @story.case(equal_to='right')
            def room_2():
                @story.part()
                def next_room_2(ctx):
                    turn = random.choice(['left', 'right'])
                    visited_rooms.receive(visited_rooms.value + 1)
                    return forking.SwitchOnValue(turn)

                @story.case(equal_to='left')
                def room_2_1():
                    @story.part()
                    def next_room_2_1(ctx):
                        visited_rooms.receive(visited_rooms.value + 1)

                @story.case(equal_to='right')
                def room_2_2():
                    @story.part()
                    def next_room_2_2(ctx):
                        visited_rooms.receive(visited_rooms.value + 1)

        await say_pure_text('enter')

        assert visited_rooms.value == 3


@pytest.mark.asyncio
async def test_one_sync_switch_inside_of_another_sync_switch_with_failed_switch():
    exit_trigger = SimpleTrigger()
    wrong_way = SimpleTrigger()

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('enter')
        def labyrinth():
            @story.part()
            def enter(ctx):
                return forking.SwitchOnValue('left')

            @story.case(equal_to='left')
            def room_1():
                @story.part()
                def next_room_1(ctx):
                    return forking.SwitchOnValue('left')

                @story.case(equal_to='left')
                def room_1_1():
                    @story.part()
                    def next_room_1_1(ctx):
                        return forking.Switch({
                            'no-exit': text.Match('no-exit'),
                        })

                    @story.case(match='no-exit')
                    def room_1_1_1():
                        @story.part()
                        def next_room_1_1_1(ctx):
                            wrong_way.passed()

                @story.part()
                def room_1_2(ctx):
                    exit_trigger.passed()
                    return [text.Any()]

        @story.on('right-way')
        def alt_labyrinth():
            @story.part()
            def next_room_2(ctx):
                wrong_way.passed()

        await say_pure_text('enter')
        # should fail switch next_room_1_1 and drop to room_1_2
        # and should not be cought by alt_labyrinth

        await say_pure_text('right-way')

        assert not wrong_way.is_triggered
        assert exit_trigger.is_triggered


@pytest.mark.asyncio
async def test_case_could_match_message_direct():
    left_trigger = SimpleTrigger()
    right_trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('open')
        def webpage_story():
            @story.part()
            async def are_you_robot(ctx):
                return await story.ask('Are you robot?', user=ctx['user'])

            @story.case(validator=text.Equal('yes'))
            def i_am_a_robot():
                @story.part()
                def left_room_passed(ctx):
                    return left_trigger.passed()

            @story.case(validator=text.Equal('no'))
            def i_am_not_a_robot():
                @story.part()
                def right_room_passed(ctx):
                    return right_trigger.passed()

        await talk.pure_text('open')
        await talk.pure_text('no')

        assert not left_trigger.is_triggered
        assert right_trigger.is_triggered


@pytest.mark.asyncio
async def test_validator_is_default_case_argument():
    left_trigger = SimpleTrigger()
    right_trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('open')
        def webpage_story():
            @story.part()
            async def are_you_robot(ctx):
                return await story.ask('Are you robot?', user=ctx['user'])

            @story.case('yes')
            def i_am_a_robot():
                @story.part()
                def left_room_passed(ctx):
                    return left_trigger.passed()

            @story.case('no')
            def i_am_not_a_robot():
                @story.part()
                def right_room_passed(ctx):
                    return right_trigger.passed()

        await talk.pure_text('open')
        await talk.pure_text('no')

        assert not left_trigger.is_triggered
        assert right_trigger.is_triggered


@pytest.mark.asyncio
@pytest.mark.parametrize('return_value, first_condition, second_condition',
                         [
                             ('right', 'left', 'right'),
                             (1, 0, 1),
                             (True, False, True),
                         ])
async def test_matche_previous_returned_value(return_value,
                                              first_condition, second_condition):
    """
    for this moment support only strings and numbers
    :return:
    """
    left_trigger = SimpleTrigger()
    right_trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('enter')
        def labyrinth():
            @story.part()
            def enter(ctx):
                return return_value

            @story.case(first_condition)
            def left_room():
                @story.part()
                def left_room_passed(ctx):
                    return left_trigger.passed()

            @story.case(second_condition)
            def right_room():
                @story.part()
                def right_room_passed(ctx):
                    return right_trigger.passed()

        await talk.pure_text('enter')

        assert not left_trigger.is_triggered
        assert right_trigger.is_triggered


@pytest.mark.asyncio
@pytest.mark.skip('test warn message here')
async def test_warn_on_incorrect_syntax_user_forgot_add_switch_value():
    left_trigger = SimpleTrigger()
    right_trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on('enter')
        def labyrinth():
            @story.part()
            def enter(ctx):
                # TODO: !!!!!!!!!!!
                # user forgot to add switch value like:
                # return 'right'
                pass

            @story.case('left')
            def left_room():
                @story.part()
                def left_room_passed(ctx):
                    return left_trigger.passed()

            @story.case('right')
            def right_room():
                @story.part()
                def right_room_passed(ctx):
                    return right_trigger.passed()

        await talk.pure_text('enter')

        assert not left_trigger.is_triggered
        assert not right_trigger.is_triggered
        # TODO: test warn message here


@pytest.mark.asyncio
async def test_one_sync_switch_inside_of_another_async_switch():
    visited_rooms = SimpleTrigger(0)

    with answer.Talk() as talk:
        story = talk.story

        @story.on('enter')
        def labyrinth():
            @story.part()
            async def enter(ctx):
                visited_rooms.receive(visited_rooms.value + 1)
                return await story.ask('Which turn to choose?', user=ctx['user'])

            @story.part()
            def parse_direction_0(ctx):
                return forking.SwitchOnValue(text.get_text(ctx)['raw'])

            @story.case(equal_to='left')
            def room_1():
                @story.part()
                async def next_room_1(ctx):
                    visited_rooms.receive(visited_rooms.value + 1)
                    return await story.ask('Which turn to choose?', user=ctx['user'])

                @story.part()
                def parse_direction_1(ctx):
                    return forking.SwitchOnValue(text.get_text(ctx)['raw'])

                @story.case(equal_to='left')
                def room_1_1():
                    @story.part()
                    def next_room_1_1(ctx):
                        visited_rooms.receive(visited_rooms.value + 1)

                @story.case(equal_to='right')
                def room_1_2():
                    @story.part()
                    def next_room_1_2(ctx):
                        visited_rooms.receive(visited_rooms.value + 1)

            @story.case(equal_to='right')
            def room_2():
                @story.part()
                async def next_room_2(ctx):
                    visited_rooms.receive(visited_rooms.value + 1)
                    return await story.ask('Which turn to choose?', user=ctx['user'])

                @story.part()
                def parse_direction_2(ctx):
                    return forking.SwitchOnValue(text.get_text(ctx)['raw'])

                @story.case(equal_to='left')
                def room_2_1():
                    @story.part()
                    def next_room_2_1(ctx):
                        visited_rooms.receive(visited_rooms.value + 1)

                @story.case(equal_to='right')
                def room_2_2():
                    @story.part()
                    def next_room_2_2(ctx):
                        visited_rooms.receive(visited_rooms.value + 1)

        await talk.pure_text('enter')
        await talk.pure_text(random.choice(['left', 'right']))
        await talk.pure_text(random.choice(['left', 'right']))

        assert visited_rooms.value == 3


@pytest.mark.asyncio
async def test_switch_inside_of_callable_inside_of_switch():
    visited_rooms = SimpleTrigger(0)
    spell_type = SimpleTrigger()
    spell_power = SimpleTrigger()

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.callable()
        def cast_the_magic():
            @story.part()
            async def ask_kind_of_spell(ctx):
                return await story.ask('What kind of spell do you cast?', user=ctx['user'])

            @story.part()
            def switch_by_kind_of_spell(ctx):
                return forking.SwitchOnValue(text.get_text(ctx)['raw'])

            @story.case(equal_to='fireball')
            def fireball():
                @story.part()
                async def power_of_spell(ctx):
                    spell_type.receive(text.get_text(ctx)['raw'])
                    return await story.ask('What is the power of fireball?', user=ctx['user'])

            @story.case(equal_to='lightning')
            def lightning():
                @story.part()
                async def power_of_spell(ctx):
                    spell_type.receive(text.get_text(ctx)['raw'])
                    return await story.ask('What is the power of lightning?', user=ctx['user'])

            @story.part()
            def store_power(ctx):
                spell_power.receive(text.get_text(ctx)['raw'])

        @story.on('enter')
        def dungeon():
            @story.part()
            async def ask_direction(ctx):
                return await story.ask(
                    'Where do you go?',
                    user=ctx['user']
                )

            @story.part()
            def parser_direction(ctx):
                return forking.SwitchOnValue(text.get_text(ctx)['raw'])

            @story.case(equal_to='left')
            def room_1():
                @story.part()
                async def meet_dragon(ctx):
                    return await cast_the_magic(session=ctx['session'], user=ctx['user'])

                @story.part()
                def store_end(ctx):
                    visited_rooms.receive(visited_rooms.value + 1)

            @story.case(equal_to='right')
            def room_2():
                @story.part()
                async def meet_ogr(ctx):
                    return await cast_the_magic(user=ctx['user'], session=ctx['session'])

                @story.part()
                def store_end(ctx):
                    visited_rooms.receive(visited_rooms.value + 1)

        await say_pure_text('enter')
        await say_pure_text(random.choice(['left', 'right']))
        await say_pure_text(random.choice(['fireball', 'lightning']))
        await say_pure_text(random.choice(['light', 'strong']))

        assert visited_rooms.value == 1
        assert spell_type.value in ['fireball', 'lightning']
        assert spell_power.value in ['light', 'strong']


@pytest.mark.asyncio
async def test_switch_without_right_case():
    got_help = SimpleTrigger()
    said_goodbay = SimpleTrigger()

    with answer.Talk() as talk:
        say_pure_text = talk(answer.pure_text)
        story = talk.story

        @story.on('I do not know')
        def meet_someone():
            @story.part()
            async def ask(message):
                return await story.ask(
                    'Do you need a help?',
                    user=message['user'],
                )

            @story.part()
            def parse_result(ctx):
                return forking.SwitchOnValue(text.get_text(ctx)['raw'])

            @story.case(equal_to='yes')
            def yes():
                @story.part()
                async def lets_go(message):
                    await story.say('Let\'s google together!', message['user'])
                    got_help.passed()

            @story.part()
            async def see_you(message):
                await story.say('Nice to see you!', user=message['user'])
                said_goodbay.passed()

        await say_pure_text('I do not know')
        await say_pure_text('No')

        assert not got_help.is_triggered
        assert said_goodbay.is_triggered


def test_serialize():
    m_old = forking.Switch({
        'location': location.Any(),
        'text': text.Any(),
    })
    ready_to_store = json.dumps(matchers.serialize(m_old))
    print('ready_to_store', ready_to_store)
    m_new = matchers.deserialize(json.loads(ready_to_store))

    assert isinstance(m_new, forking.Switch)
    assert isinstance(m_new.cases['location'], type(m_old.cases['location']))
    assert isinstance(m_new.cases['text'], type(m_old.cases['text']))



import logging
import inspect

from . import parser

logger = logging.getLogger(__name__)


class Scope:
    def __init__(self, name, parent=None):
        self.storage = {}
        # instances that will auto update on each new instance come
        self.auto_bind_list = []
        # description of classes
        self.described = {}
        # functions that waits for deps
        self.requires_fns = {}
        # all instances that are singletones
        self.singleton_cache = {}
        self.name = name
        # parent scope
        self.parent = parent

    def get(self, type_name):
        try:
            item = self.storage[type_name]
            if inspect.isclass(item):
                return item()
            else:
                return item
        except KeyError as err:
            if self.parent:
                return self.parent.get(type_name)
            raise err

    def get_instance(self, type_name):
        try:
            return self.singleton_cache[type_name]
        except KeyError as err:
            if self.parent:
                return self.parent.get_instance(type_name)
            raise err

    def describe(self, type_name, cls):
        self.described[cls] = {
            'type': type_name,
        }

    def get_auto_bind_list(self):
        yield from self.auto_bind_list
        if self.parent:
            yield from self.parent.get_auto_bind_list()

    def auto_bind(self, instance):
        self.auto_bind_list.append(instance)

    def get_description(self, value):
        try:
            return self.described.get(value, self.described[type(value)])
        except KeyError as err:
            if self.parent:
                return self.parent.get_description(value)
            raise err

    def store_instance(self, type_name, instance):
        self.singleton_cache[type_name] = instance

    def get_endpoint_deps(self, method_ptr):
        return self.requires_fns.get(method_ptr, {}).items() or \
               self.parent and self.parent.get_endpoint_deps(method_ptr) or \
               []

    def store_deps_endpoint(self, fn, deps):
        self.requires_fns[fn] = deps

    def clear(self):
        self.auto_bind_list = []
        self.described = {}
        self.requires_fns = {}
        self.singleton_cache = {}

    def clear_instances(self):
        self.auto_bind_list = []
        self.singleton_cache = {}
        self.storage = {}

    def register(self, type_name, value):
        if type_name in self.storage:
            self.remove_type(type_name)

        self.storage[type_name] = value

    def remove_type(self, type_name):
        self.storage.pop(type_name, True)
        instance = self.singleton_cache.pop(type_name, True)
        # TODO: clear self.requires_fns and self.auto_update_list maybe we can use type(instance)?

    def __repr__(self):
        return '<Scope {}> {}'.format(self.name, {
            'storage': self.storage,
            'auto_update_list': self.auto_bind_list,
            'described': self.described,
            'requires_fns': self.requires_fns,
            'singleton_cache': self.singleton_cache,
        })


def empty_array_if_empty(default):
    return [default] if default is not inspect.Parameter.empty else []


class MissedDescriptionError(Exception):
    pass


class Injector:
    def __init__(self):
        self.root = Scope('root')
        self.current_scope = self.root

    def describe(self, type_name, cls):
        """
        add description of class

        :param type_name:
        :param cls:
        :return:
        """

        self.current_scope.describe(type_name, cls)

    def register(self, type_name=None, instance=None):
        print()
        print('register {} = {}'.format(type_name, instance))
        if not isinstance(type_name, str) and type_name is not None:
            raise ValueError('type_name parameter should be string or None')
        if type_name is None:
            try:
                desc = self.current_scope.get_description(instance)
            except KeyError:
                # TODO: should raise exception
                # raise MissedDescriptionError('{} was not registered'.format(instance))
                # print('self.described')
                # print(self.current_scope.described)
                # print('type(instance)')
                # print(type(instance))
                # print('self.described.get(type(instance))')
                # print(self.current_scope.described.get(type(instance)))
                print('{} was not registered'.format(instance))
                return None
            type_name = desc['type']
        # print('> before store {} = {}'.format(type_name, instance))
        self.current_scope.register(type_name, instance)
        for wait_instance in self.current_scope.get_auto_bind_list():
            self.bind(wait_instance)

    def requires(self, fn):
        fn_sig = inspect.signature(fn)
        self.current_scope.store_deps_endpoint(fn, {
            key: {'default': default for default in empty_array_if_empty(fn_sig.parameters[key].default)}
            for key in fn_sig.parameters.keys() if key != 'self'})

    def entrypoint_deps(self, method_ptr):
        # we should have something to inject
        # registered instance, class or default value
        if not any(self.get(dep) or 'default' in dep_spec
                   for dep, dep_spec
                   in self.current_scope.get_endpoint_deps(method_ptr)):
            # otherwise we should inject anything
            return {}

        return {dep: self.get(dep) or dep_spec['default']
                for dep, dep_spec in self.current_scope.get_endpoint_deps(method_ptr)}

    def bind(self, instance, auto=False):
        """
        Bind deps to instance

        :param instance:
        :param auto: follow update of DI and refresh binds once we will get something new
        :return:
        """
        methods = [
            (m, cls.__dict__[m])
            for cls in inspect.getmro(type(instance))
            for m in cls.__dict__ if inspect.isfunction(cls.__dict__[m])
            ]

        try:
            deps_of_endpoints = [(method_ptr, self.entrypoint_deps(method_ptr))
                                 for (method_name, method_ptr) in methods]

            for (method_ptr, method_deps) in deps_of_endpoints:
                if len(method_deps) > 0:
                    method_ptr(instance, **method_deps)
        except KeyError:
            pass

        if auto and instance not in self.current_scope.get_auto_bind_list():
            self.current_scope.auto_bind(instance)

        return instance

    def get(self, type_name):
        type_name = parser.kebab_to_underscore(type_name)
        try:
            return self.current_scope.get_instance(type_name)
        except KeyError:
            try:
                instance = self.current_scope.get(type_name)
            except KeyError:
                # TODO: sometimes we should fail loudly in this case
                return None

            self.current_scope.store_instance(type_name, instance)
            instance = self.bind(instance)

        return instance

    def child_scope(self, name='undefined'):
        return ChildScopeBuilder(self, self.current_scope, name)

    def clear_instances(self):
        self.current_scope.clear_instances()

    def add_scope(self, scope):
        self.current_scope = scope

    def remove_scope(self, scope):
        assert self.current_scope == scope
        self.current_scope = scope.parent


class ChildScopeBuilder:
    def __init__(self, injector, parent, name):
        self.injector = injector
        self.name = name
        self.parent = parent

    def __enter__(self):
        self.scope = Scope(self.name, self.parent)
        self.injector.add_scope(self.scope)
        return self.scope

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.scope.clear()
        self.injector.remove_scope(self.scope)



from botstory import ast, matchers
from botstory.ast import forking
import logging
import json

logger = logging.getLogger(__name__)


# 1. once we trap on StoriesLoopNode we should stop execution
# and wait any user input
# 2. once we got any user import we should come to StoriesLoopNode
# add check whether input match received request.
# 3. execute matched story


class BreakLoop:
    """
    break outsize of story loop
    """
    type = 'BreakLoop'


class StoryLoopAPI:
    """
    loop (scope) concept similar to switch (forking)
    the main difference is that after case switch jump to the next part
    after last case. But in scope it loop until
    - we get receive unmatched message
    - or break loop explicitly
    """

    def __init__(self, library, parser_instance):
        self.library = library
        self.parser_instance = parser_instance

    def loop(self):
        def fn(one_loop):
            scope_node = StoriesLoopNode(one_loop)
            self.parser_instance.add_to_current_node(scope_node)
            self.parser_instance.compile_scope(scope_node, one_loop)
            return one_loop

        return fn


class StoriesLoopNode:
    def __init__(self, target):
        self.local_scope = ast.library.StoriesScope()
        self.story_line = []
        self.target = target
        self.topic = target.__name__

    @property
    def __name__(self):
        return self.target.__name__

    def by_topic(self, topic):
        stories = self.local_scope.by_topic(topic)
        return stories[0] if len(stories) > 0 else None

    @property
    def children(self):
        return self.local_scope.stories

    def children_matcher(self):
        return ScopeMatcher(forking.Switch(self.local_scope.all_filters()))

    def get_child_by_validation_result(self, topic):
        case_stories = self.local_scope.by_topic(topic)

        if len(case_stories) == 0:
            logger.debug('#######################################')
            logger.debug('# [!] do not have any child here')
            logger.debug('# story node = {}'.format(self))
            logger.debug('# topic = {}'.format(topic))
            logger.debug('#######################################')
            return None

        return case_stories[0]

    def match_children(self, key, value):
        return [child for child in self.local_scope.stories
                if child.extensions.get(key, forking.Undefined) == value]

    def to_json(self):
        return {
            'type': 'StoriesLoopNode',
            'name': self.__name__,
            'local_scope': self.local_scope.to_json()
        }

    def __repr__(self):
        return json.dumps(self.to_json())


@matchers.matcher()
class ScopeMatcher:
    def __init__(self, all_filters):
        # TODO: It's not good name
        # but better what we have for this moment
        # so it could be renamed very soon
        self.new_scope = True
        self.all_filters = all_filters

    def validate(self, message):
        return self.all_filters.validate(message)

    def serialize(self):
        return self.all_filters.serialize()

    @classmethod
    def deserialize(cls, data):
        return cls(forking.Switch.deserialize(data))



import re


def camel_case_to_underscore(class_name):
    """Converts normal class names into normal arg names.
    Normal class names are assumed to be CamelCase with an optional leading
    underscore.  Normal arg names are assumed to be lower_with_underscores.
    Args:
      class_name: a class name, e.g., "FooBar" or "_FooBar"
    Returns:
      all likely corresponding arg names, e.g., ["foo_bar"]

    based on: <https://github.com/google/pinject/blob/master/pinject/bindings.py>

    """
    parts = []
    rest = class_name
    if rest.startswith('_'):
        rest = rest[1:]
    while True:
        m = re.match(r'([A-Z][a-z]+)(.*)', rest)
        if m is None:
            break
        parts.append(m.group(1))
        rest = m.group(2)
    if not parts:
        return []
    return ['_'.join(part.lower() for part in parts)]


def kebab_to_underscore(s):
    """
    Convert kebab-styled-string to underscore_styled_string
    :param s:
    :return:
    """
    return s.replace('-', '_')



import pytest
from . import messenger
from .. import mockhttp
from ... import utils


@pytest.mark.parametrize('greeting_text,valid', [
    ('very-long-message ' * 100, False),
    ('short-message.', True),
])
@pytest.mark.asyncio
async def test_validate_greeting_text(mocker, greeting_text, valid):
    fb = messenger.FBInterface()
    log_mock = mocker.patch('botstory.integrations.fb.messenger.logger')
    fb.http = mockhttp.MockHttpInterface()
    await fb.set_greeting_text(greeting_text)
    if valid:
        assert not log_mock.warn.called, 'should be valid'
    else:
        log_mock.warn.assert_called_with('greeting text should not exceed 160 length in characters')


@pytest.mark.parametrize('menu,invalid_message', [
    ([{
        'type': 'postback',
        'title': 'Help',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }] * 2, False),

    ([{
        'type': 'postback',
        'title': 'Help',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }] * 10, 'menu should not exceed 5 call to actions'),

    ([{
        'type': 'postback',
        'title': 'Help',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }, {
        'type': 'postback',
        'title': 'VeryLongTitle' * 10,
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }], 'menu item title should not exceed 30 characters'),

    ([{
        'type': 'postback',
        'title': 'Help',
        'payload': 'VERY_LONG_PAYLOAD_' * 100
    }, {
        'type': 'postback',
        'title': 'Hello',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }], 'menu item payload should not exceed 1000 characters'),
])
@pytest.mark.asyncio
async def test_validate_persistent_menu(mocker, menu, invalid_message):
    fb = messenger.FBInterface()
    log_mock = mocker.patch('botstory.integrations.fb.messenger.logger')
    fb.http = mockhttp.MockHttpInterface()
    await fb.set_persistent_menu(menu)
    if invalid_message:
        log_mock.warn.assert_called_with(invalid_message)
    else:
        assert not log_mock.warn.called, 'should be valid'


@pytest.mark.parametrize('text,options,invalid_message', [
    ('hi there!', None, False),

    ('very long message ' * 40, None, 'send message text should not exceed 640 character limit'),

    ('short message', [{
        'content_type': 'text',
        'title': 'Red',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_PICKING_RED',
    }] * 20, 'send message quick replies should not exceed 10 limit'),

    ('short message', [{
        'content_type': 'text',
        'title': 'Very Long ' * 10,
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_PICKING_RED',
    }], 'send message quick replies title should not exceed 20 character limit'),

    ('short message', [{
        'content_type': 'text',
        'title': 'Short',
        'payload': 'VERY_LONG_' * 200,
    }], 'send message quick replies payload should not exceed 1000 character limit'),

    ('short message', [{
        'title': 'Short',
        'payload': 'SOME_PAYLOAD',
    }], 'send message quick replies should have content_type'),
])
@pytest.mark.asyncio
async def test_validate_send_text_message(mocker, text, options, invalid_message):
    recipient = utils.build_fake_user()
    fb = messenger.FBInterface()
    log_mock = mocker.patch('botstory.integrations.fb.messenger.logger')
    fb.http = mockhttp.MockHttpInterface()
    await fb.send_text_message(recipient, text, options)
    if invalid_message:
        log_mock.warn.assert_called_with(invalid_message)
    else:
        assert not log_mock.warn.called, 'should be valid'



from botstory.middlewares.sticker.sticker import get_sticker, Any, Like, \
    ANY_LIKE, SMALL_LIKE, MEDIUM_LIKE, BIG_LIKE, LIKE_STICKERS

__all__ = [get_sticker, Any, Like, ANY_LIKE, SMALL_LIKE, MEDIUM_LIKE, BIG_LIKE, LIKE_STICKERS]



class Invalid(BaseException):
    pass


class ExceedLengthException(Invalid):
    def __init__(self, *args, limit=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.limit = limit


def greeting_text(message):
    """
    more: https://developers.facebook.com/docs/messenger-platform/thread-settings/greeting-text

    :param message:
    :return:
    """
    if len(message) > 160:
        raise Invalid('greeting text should not exceed 160 length in characters')


def persistent_menu(menu):
    """
    more: https://developers.facebook.com/docs/messenger-platform/thread-settings/persistent-menu

    :param menu:
    :return:
    """
    if len(menu) > 5:
        raise Invalid('menu should not exceed 5 call to actions')

    for item in menu:
        if len(item['title']) > 30:
            raise Invalid('menu item title should not exceed 30 characters')

        if item['type'] == 'postback' and len(item['payload']) > 1000:
            raise Invalid('menu item payload should not exceed 1000 characters')


def send_text_message(text, quick_replies):
    """
    more:
    https://developers.facebook.com/docs/messenger-platform/send-api-reference/text-message
     and
    https://developers.facebook.com/docs/messenger-platform/send-api-reference/quick-replies

    :param text:
    :param quick_replies:
    :return:
    """
    if len(text) > 640:
        raise ExceedLengthException(
            'send message text should not exceed 640 character limit',
            limit=640,
        )

    if isinstance(quick_replies, list):
        if len(quick_replies) > 10:
            raise Invalid('send message quick replies should not exceed 10 limit')

        for item in quick_replies:
            if len(item['title']) > 20:
                raise Invalid('send message quick replies title should not exceed 20 character limit')
            if 'content_type' not in item:
                raise Invalid('send message quick replies should have content_type')
            if item['content_type'] == 'text' and len(item['payload']) > 1000:
                raise Invalid('send message quick replies payload should not exceed 1000 character limit')



from botstory import di
import logging
import json

logger = logging.getLogger(__name__)


@di.desc(reg=False)
class Parser:
    def __init__(self, library):
        self.current_node = None
        self.current_scope = library.global_scope

    def compile(self, one_story):
        topic = one_story.__name__
        previous_node = self.current_node
        self.current_node = ASTNode(topic=topic)

        one_story()

        res = self.current_node
        self.current_node = previous_node
        return res

    def compile_fork(self, fork_node, one_story):
        parent_scope = self.current_scope
        self.current_scope = fork_node.local_scope

        compiled_story = self.compile(one_story)

        self.current_scope.add(compiled_story)

        self.current_scope = parent_scope

        return compiled_story

    def compile_scope(self, scope_node, scope_func):
        parent_scope = self.current_scope
        self.current_scope = scope_node.local_scope

        scope_func()

        res = self.current_scope
        self.current_scope = parent_scope
        return res

    def get_last_story_part(self):
        return self.current_node.story_line[-1] \
            if len(self.current_node.story_line) > 0 else None

    def add_to_current_node(self, node):
        self.current_node.append(node)

    def part(self, story_part):
        self.current_node.append(story_part)
        return True

    @property
    def topic(self):
        return self.current_node.topic


class ASTNode:
    def __init__(self, topic):
        self.compiled_story = None
        self.extensions = {}
        self.story_line = []
        self.story_names = set()
        self.topic = topic

    def add_child(self, child_story_line):
        """
        add child node to the last part of story
        :param child_story_line:
        :return:
        """
        # assert isinstance(self.story_line[-1], StoryPartFork)
        self.story_line[-1].add_child(child_story_line)

    def append(self, story_part):
        part_name = story_part.__name__,
        if part_name in self.story_names:
            logger.warning('Already have story with name {}. Please use uniq name'.format(part_name))

        self.story_names.add(part_name)
        self.story_line.append(story_part)
        return story_part

    def to_json(self):
        return {
            'type': 'ASTNode',
            'extensions': str(self.extensions),
            'topic': self.topic,
            'story_line': list(
                [l.to_json() if hasattr(l, 'to_json') else 'part: {}'.format(l.__name__) for l in self.story_line]
            ),
        }

    def __repr__(self):
        return json.dumps(self.to_json())


class StoryPartLeaf:
    def __init__(self, fn):
        self.fn = fn

    @property
    def __name__(self):
        return self.fn.__name__

    def __call__(self, *args, **kwargs):
        # TODO: because we could have async and sync parts
        # we should check it here
        return self.fn(*args, **kwargs)

    def __repr__(self):
        return json.dumps({
            'type': 'StoryPartLeaf',
            'name': self.__name__,
        })



from botstory import di
from botstory.ast import forking
import logging
import json

logger = logging.getLogger(__name__)


class DuplicationException(Exception):
    pass


class StoriesScope:
    def __init__(self):
        self.stories = []

    def add(self, story):
        if self.by_topic(story.topic):
            raise DuplicationException('We already have topic {}'.format(story.topic))
        self.stories.append(story)

    def all_filters(self):
        return {s.topic: s.extensions['validator']
                for s in self.stories if 'validator' in s.extensions}

    def match(self, message):
        matched_stories = [
            story for story in self.stories
            if story.extensions['validator'].validate(message)]
        return matched_stories[0] if len(matched_stories) > 0 else None

    def get_story_by(self, **kwargs):
        return [child for child in self.stories
                if all(child.extensions.get(key, forking.Undefined) == kwargs[key]
                       for key in kwargs.keys())]

    def by_topic(self, topic):
        return [s for s in self.stories if s.topic == topic]

    def to_json(self):
        return {
            'type': 'StoriesScope',
            'stories': [s.to_json() for s in self.stories]
        }

    def __repr__(self):
        return json.dumps(self.to_json())


@di.desc(reg=False)
class StoriesLibrary:
    """
    storage of
     - global scope stories
     - callable stories
    """

    def __init__(self):
        self.callable_scope = StoriesScope()
        self.global_scope = StoriesScope()

    def clear(self):
        self.callable_scope = StoriesScope()
        self.global_scope = StoriesScope()

    def add_global(self, story):
        self.global_scope.add(story)

    def add_callable(self, story):
        self.callable_scope.add(story)

    def get_callable_by_topic(self, topic):
        return self.callable_scope.by_topic(topic)[0]

    def get_global_story(self, message):
        return self.global_scope.match(message)

    def get_story_by_topic(self, topic, stack=None):
        """
        get story and take about context stack
        :param topic:
        :param stack:
        :return:
        """
        options = self.callable_scope.by_topic(topic) + self.global_scope.by_topic(topic)

        if len(options) > 0:
            return options[0]

        if not stack or len(stack) == 0:
            return None

        # Or it seems that we are somewhere on a leaves
        # so we're trying to find root of last story and than
        # get right sub story

        parent = self.get_story_by_topic(stack[-1]['topic'], stack[:-1])
        if not parent:
            return None

        # is topic name matching storyline?
        try:
            return next(filter(lambda part: getattr(part, 'topic', None) == topic, parent.story_line))
        except StopIteration:
            pass

        if hasattr(parent, 'local_scope'):
            # for loop.StoriesLoopNode
            return parent.get_child_by_validation_result(topic)

        # for forking.StoryPartFork
        inner_stories = [
            story.children for story in parent.story_line if hasattr(story, 'children')
        ]

        inner_stories = [item for sublist in inner_stories for item in sublist]

        child_options = [story for story in inner_stories if story.topic == topic]
        if len(child_options) == 0:
            return None
        elif len(child_options) == 1:
            return child_options[0]
        else:
            raise Exception('We have few options with the same name {}'.format(topic))



import aiohttp
from aiohttp.helpers import sentinel
import json
from unittest import mock

from ... import di


def stub(name=None):
    """
    Creates a stub method. It accepts any arguments. Ideal to register to
    callbacks in tests.
    :param name: the constructed stub's name as used in repr
    :rtype: mock.MagicMock
    :return: Stub object.
    """
    return mock.MagicMock(spec=lambda *args, **kwargs: None, name=name)


@di.desc('http', reg=False)
class MockHttpInterface:
    def __init__(self,
                 get={}, get_raise=sentinel,
                 post=True, post_raise=sentinel,
                 put=True, put_raise=sentinel,
                 delete=True, delete_raise=sentinel,
                 start=True,
                 stop=True):
        self.get = aiohttp.test_utils.make_mocked_coro(
            return_value=get,
            raise_exception=get_raise,
        )
        self.get_raw = aiohttp.test_utils.make_mocked_coro(
            return_value={'text': json.dumps(get), 'status': 200, },
            raise_exception=get_raise,
        )
        self.post = aiohttp.test_utils.make_mocked_coro(
            return_value=post,
            raise_exception=post_raise,
        )
        self.post_raw = aiohttp.test_utils.make_mocked_coro(
            return_value={'text': json.dumps(post), 'status': 200, },
            raise_exception=post_raise,
        )
        self.delete = aiohttp.test_utils.make_mocked_coro(
            return_value=delete,
            raise_exception=delete_raise,
        )
        self.put = aiohttp.test_utils.make_mocked_coro(
            return_value=put,
            raise_exception=put_raise,
        )
        self.setup = aiohttp.test_utils.make_mocked_coro()
        self.start = aiohttp.test_utils.make_mocked_coro(return_value=start)
        self.stop = aiohttp.test_utils.make_mocked_coro(return_value=stop)
        self.webhook = stub('webhook')



import os

from setuptools import setup, find_packages

longDesc = ''
if os.path.exists('README.md'):
    longDesc = open('README.md').read().strip()

version = ''
if os.path.exists('version.txt'):
    version = open('version.txt').read().strip()

setup(
    name='botstory',
    packages=find_packages(),
    version=version,
    description='Async framework for bots',
    license='MIT',
    long_description=longDesc,
    author='Eugene Krevenets',
    author_email='ievegenii.krevenets@gmail.com',
    url='https://github.com/botstory/bot-story',
    download_url='https://github.com/botstory/bot-story/tarball/{}'.format(version),
    keywords=['bot', 'ai', 'nlp', 'asyncio'],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Topic :: Communications :: Chat',

        # Not early because of async/await
        'Programming Language :: Python :: 3.5',
    ],
    install_requires=[
        'aiohttp==1.1.6',
        'motor==1.0',
        'yarl==0.7.1',
    ],
    package_data={
        '': ['*.crt', '*.key']
    },
)



from botstory import matchers, utils
from botstory.ast import callable, forking, loop
from botstory.ast.story_context import reducers
from botstory.utils import advanced_json_encoder
import numbers

import logging
import uuid

logger = logging.getLogger(__name__)


class MissedStoryPart(Exception):
    pass


class StoryContext:
    def __init__(self, message, library, matched=False, waiting_for=None, parent_uid=None):
        self.uid = str(uuid.uuid4())
        self.parent_uid = parent_uid
        self.library = library
        # whether message was passed validation and was matched one story
        self.matched = matched
        self.message = message
        self.waiting_for = waiting_for

    def clone(self):
        return StoryContext(library=self.library,
                            matched=self.matched,
                            message=self.message,
                            parent_uid=self.uid,
                            waiting_for=self.waiting_for,
                            )

    def compiled_story(self):
        if self.is_empty_stack():
            return self.library.get_global_story(self.message)
        else:
            return self.library.get_story_by_topic(self.stack_tail()['topic'], stack=self.stack()[:-1])

    def could_scope_out(self):
        """
        could bubble up from current scope

        :return:
        """
        return not self.waiting_for or \
               isinstance(self.waiting_for, callable.EndOfStory) or \
               self.is_breaking_a_loop()

    def current_step(self):
        return self.stack_tail()['step']

    def does_it_match_any_story(self):
        return self.compiled_story() is not None and not self.matched

    def get_child_story(self):
        logger.debug('get_child_story')
        """
        try child story that match message and get scope of it
        :return:
        """
        story_loop = self.compiled_story()
        if hasattr(story_loop, 'children_matcher') and not self.matched:
            return self.get_story_scope_child(story_loop)

        story_part = self.get_current_story_part()

        if not hasattr(story_part, 'get_child_by_validation_result'):
            return None

        if isinstance(self.waiting_for, forking.SwitchOnValue):
            return story_part.get_child_by_validation_result(self.waiting_for.value)

        # for some base classes we could try validate result direct
        child_story = story_part.get_child_by_validation_result(self.waiting_for)
        logger.debug('child_story')
        logger.debug(child_story)
        if child_story:
            return child_story

        stack_tail = self.stack_tail()
        if stack_tail['data'] is not None and not self.matched:
            validator = matchers.deserialize(stack_tail['data'])
            validation_result = validator.validate(self.message)
            res = story_part.get_child_by_validation_result(validation_result)
            # or we validate message
            # but can't find right child story
            # maybe we should use independent validators for each story here
            if res is None:
                return self.get_story_scope_child(story_part)
            else:
                return res

        return None

    def get_story_scope_child(self, story_part):
        validator = story_part.children_matcher()
        topic = validator.validate(self.message)
        # if topic == None:
        # we inside story loop scope
        # but got message that doesn't match
        # any local stories
        return story_part.by_topic(topic)

    def get_current_story_part(self):
        compiled_story = self.compiled_story()
        if not compiled_story:
            return None
        try:
            return compiled_story.story_line[self.current_step()]
        except IndexError:
            return None

    def get_user_data(self):
        return get_user_data(self.message)

    def has_child_story(self):
        return self.get_child_story() is not None

    def is_breaking_a_loop(self):
        return isinstance(self.waiting_for, loop.BreakLoop)

    def is_empty_stack(self):
        return len(self.stack()) == 0

    def is_end_of_story(self):
        compiled_story = self.compiled_story()
        if not compiled_story:
            return True
        return self.current_step() >= len(compiled_story.story_line)

    def is_scope_level(self):
        return isinstance(self.compiled_story(), loop.StoriesLoopNode)

    def is_scope_level_part(self):
        return isinstance(self.get_current_story_part(), loop.StoriesLoopNode)

    def is_tail_of_story(self):
        compiled_story = self.compiled_story()
        if not compiled_story:
            return True
        return self.current_step() >= len(compiled_story.story_line) - 1

    def is_waiting_for_input(self):
        """
        could make one step further
        :return:
        """
        return self.waiting_for and \
               not isinstance(self.waiting_for, forking.SwitchOnValue) and \
               not is_base_type(self.waiting_for)

    def stack(self):
        return self.message['session']['stack']

    def stack_tail(self):
        stack = self.stack()
        if len(stack) == 0:
            raise MissedStoryPart()
        return stack[-1]

    def to_json(self):
        return {
            'uid': self.uid,
            'parent_uid': self.parent_uid,
            'matched': self.matched,
            'message': self.message,
            'waiting_for': str(self.waiting_for),
        }

    def user(self):
        return self.message['user']

    def __repr__(self):
        try:
            return advanced_json_encoder.AdvancedJSONEncoder().encode(self.to_json())
        except Exception as err:
            logger.warn(err)
            logger.warn('fail to dump json of message {} '
                        'waiting for {}'.format(str(self.message), str(self.waiting_for)))


def is_base_type(value):
    return isinstance(value, str) or isinstance(value, numbers.Number)


def clean_message_data(ctx):
    return utils.safe_set(ctx, 'session', 'data', 'message', {})


def get_user_data(ctx):
    return ctx['session']['data']


def get_message_data(ctx, *args, **kwargs):
    return utils.safe_get(get_user_data(ctx)['message'], *args, **kwargs)


def set_user_data(ctx, data):
    ctx['session']['data'] = {
        **get_user_data(ctx),
        **data,
    }
    return ctx


def set_message_data(ctx, *args):
    return utils.safe_set(ctx, 'session', 'data', 'message', *args)


__all__ = [get_user_data, reducers, set_user_data, StoryContext]



import inspect
from .parser import camel_case_to_underscore
from .. import di


def desc(t=None, reg=True):
    """
    Describe Class Dependency

    :param reg: should we register this class as well
    :param t: custom type as well
    :return:
    """

    def decorated_fn(cls):
        if not inspect.isclass(cls):
            return NotImplemented('For now we can only describe classes')
        name = t or camel_case_to_underscore(cls.__name__)[0]
        if reg:
            di.injector.register(name, cls)
        else:
            di.injector.describe(name, cls)
        return cls

    return decorated_fn



import pytest
from . import location
from ... import matchers
from ...utils import answer, SimpleTrigger


@pytest.mark.asyncio
async def test_should_trigger_on_any_location(event_loop):
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(receive=location.Any())
        def one_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        assert not event_loop.is_closed()
        await story.start(event_loop)

        await talk.location({'lat': 1, 'lng': 1})
        assert trigger.is_triggered


@pytest.mark.asyncio
async def test_should_not_react_on_common_message():
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story

        @story.on(receive=location.Any())
        def one_story():
            @story.part()
            def then(ctx):
                trigger.passed()

        await talk.pure_text('Hey!')

        assert not trigger.is_triggered


def test_serialize_location():
    m_old = location.Any()
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, location.Any)



import logging
from .. import matchers, middlewares

logger = logging.getLogger(__name__)


class CommonStoriesAPI:
    def __init__(self, parser, library):
        self.parser = parser
        self.library = library

    def on(self, receive):
        def fn(one_story):
            compiled_story = self.parser.compile(
                one_story,
            )
            compiled_story.extensions['validator'] = matchers.get_validator(receive)
            self.parser.current_scope.add(compiled_story)

            return one_story

        return fn

    def on_start(self):
        def fn(one_story):
            compiled_story = self.parser.compile(
                one_story,
            )
            compiled_story.extensions['validator'] = middlewares.option.OnStart()
            self.library.add_global(compiled_story)

            return one_story

        return fn

    def part(self):
        def fn(part_of_story):
            self.parser.part(part_of_story)
            return part_of_story

        return fn



from .parser import camel_case_to_underscore, kebab_to_underscore


def test_camelcase_to_underscore():
    assert camel_case_to_underscore('ClassName')[0] == 'class_name'


def test_remove_leading_underscore():
    assert camel_case_to_underscore('_ClassName')[0] == 'class_name'


def test_should_return_empty_array_if_no_any_class_name_here():
    assert camel_case_to_underscore('_qwerty') == []


def test_kebab_to_underscore():
    assert kebab_to_underscore('hello-world') == 'hello_world'



import json
import logging
logger = logging.getLogger(__name__)


class JSDict:
    def __init__(self, response):
        # don't use self.__dict__ here
        self._response = response

    def __repr__(self):
        return self.__str__()

    def __getattr__(self, key):
        return self._response.get(key, None)

    def __getitem__(self, item):
        return self._response.get(item, None)

    def __str__(self):
        return 'JSDict({})'.format(json.dumps(self._response))

    def __iter__(self):
        logger.debug('JSDict.__iter__')
        logger.debug(list(iter(self._response)))
        return iter(self._response)

    def items(self):
        return self._response.items()



"""
inspired by
https://github.com/KeepSafe/aiohttp/blob/master/examples/fake_server.py

"""

from aiohttp import web
import logging

from ...tests.fake_server import FakeServer, get, post

logger = logging.getLogger(__name__)


class FakeAnalytics(FakeServer):
    ROOT_URI = 'www.google-analytics.com'

    @get('/collect')
    async def on_me(self, request):
        return web.json_response({})

    @post('/collect')
    async def on_my_friends(self, request):
        return web.json_response({})



import asyncio
import random

from .jsdict import JSDict


async def a_second():
    await asyncio.sleep(0.1)


def uniq_id():
    return random.randint(100000000, 123456789)


def build_fake_user():
    return {
        '_id': uniq_id(),
        'name': 'Alice',
        'slack_id': uniq_id(),
        'facebook_user_id': uniq_id(),
    }


def build_fake_session(user=None):
    user = user or {}
    return {
        'data': {},
        'stack': [],
        'facebook_user_id': user.get('facebook_user_id', uniq_id()),
        'user_id': user.get('_id', uniq_id()),
    }


def safe_get(dct, *keys, default=None):
    if dct is None:
        return None

    for key in keys:
        try:
            dct = dct[key]
        except (KeyError, TypeError):
            return default
    return dct


def safe_set(input_dict, *args):
    res = input_dict
    keys, last_key, value = args[:-2], args[-2], args[-1]
    keys_iter = iter(keys)
    try:
        key = None
        try:
            while True:
                key = next(keys_iter)
                res = res[key]
        except (KeyError, TypeError):
            while True:
                res[key] = {}
                res = res[key]
                key = next(keys_iter)
    except StopIteration:
        pass

    res[last_key] = value
    return input_dict


class SimpleTrigger:
    def __init__(self, value=None):
        self.is_triggered = False
        self.triggered_times = 0
        self.value = value

    def passed(self):
        self.is_triggered = True
        self.triggered_times += 1

    def is_passed(self):
        return self.is_triggered

    def inc(self):
        self.value += 1

    def receive(self, value):
        self.value = value

    def result(self):
        return self.value


def uniqify(seq, idfun=None):
    """
    remove duplicates

    get here https://www.peterbe.com/plog/uniqifiers-benchmark

    :param seq:
    :param idfun:
    :return:
    """
    # order preserving
    if idfun is None:
        def idfun(x): return x
    seen = {}
    result = []
    for item in seq:
        marker = idfun(item)
        # in old Python versions:
        # if seen.has_key(marker)
        # but in new ones:
        if marker in seen: continue
        seen[marker] = 1
        result.append(item)
    return result


def is_string(obj):
    if isinstance(obj, str) or isinstance(obj, bytes) or isinstance(obj, bytearray):
        return True
    return False



"""
pageview: [ page path ]
event: category, action, [ label [, value ] ]
social: network, action [, target ]
timing: category, variable, time [, label ]
"""

from botstory.ast import story_context
import functools
import json
import logging
from .universal_analytics.tracker import Tracker

from ... import di
from ...utils import queue

logger = logging.getLogger(__name__)


@di.desc('tracker', reg=False)
class GAStatistics:
    def __init__(self,
                 tracking_id=None,
                 story_tracking_template='{story}/{part}',
                 new_message_tracking_template='receive: {data}',
                 ):
        """
        :param tracking_id: should be like UA-XXXXX-Y
        """
        self.tracking_id = tracking_id
        self.story_tracking_template = story_tracking_template
        self.new_message_tracking_template = new_message_tracking_template

    @staticmethod
    def __hash__():
        return hash('ga.tracker')

    def get_tracker(self, user):
        logger.debug('get_tracker')
        logger.debug(Tracker)
        return Tracker(
            account=self.tracking_id,
            client_id=user and user['_id'],
        )

    def event(self, user,
              event_category=None,
              event_action=None,
              event_label=None,
              event_value=None,
              ):
        queue.add(
            functools.partial(self.get_tracker(user).send,
                              'event', event_category, event_action, event_label, event_value
                              )
        )

    def story(self, ctx):
        user = ctx.user()
        story_name = ctx.stack()[-1]['topic']
        story_part_name = ctx.get_current_story_part().__name__

        queue.add(
            functools.partial(self.get_tracker(user).send,
                              'pageview', self.story_tracking_template.format(story=story_name,
                                                                              part=story_part_name),
                              )
        )

    def new_message(self, ctx):
        user = ctx.message['user']
        data = ctx.get_user_data()
        queue.add(
            functools.partial(self.get_tracker(user).send,
                              'pageview', self.new_message_tracking_template.format(data=json.dumps(data)),
                              )
        )

    def new_user(self, user):
        queue.add(
            functools.partial(self.get_tracker(user).send,
                              'event',
                              'new_user', 'start', 'new user starts chat'
                              )
        )
